from .encrypt import encrypt, decrypt
from .Riddle import Riddle
from .Locket import Locket
from .SecretKeeper import SecretKeeper
from .SecretKeeper import Anonymizer as ListAnonymizer
from .SecretKeeper import SecretKeeper as DataAnonymizer




import base64

def encrypt(key, clear):
	enc = []
	for i in range(len(clear)):
		key_c = key[i % len(key)]
		enc_c = chr((ord(clear[i]) + ord(key_c)) % 256)
		enc.append(enc_c)
	return base64.urlsafe_b64encode("".join(enc).encode()).decode()

def decrypt(key, encrypted):
	dec = []
	encrypted = base64.urlsafe_b64decode(encrypted).decode()
	for i in range(len(encrypted)):
		key_c = key[i % len(key)]
		dec_c = chr((256 + ord(encrypted[i]) - ord(key_c)) % 256)
		dec.append(dec_c)
	return "".join(dec)



from chronometry.progress import ProgressBar, iterate
import secrets
import random
from pandas import DataFrame, Series


def get_nonzero_random_below(n):
	if secrets.randbelow(2) == 0:
		return secrets.randbelow(n-1) + 1
	else:
		return -secrets.randbelow(n-1) - 1


class Anonymizer:
	def __init__(self, map_numbers, upper_bound=1000):
		self._values_to_keys = {}
		self._keys_to_values = {}
		self._count = 0
		self._coefficient = get_nonzero_random_below(upper_bound)
		self._constant = get_nonzero_random_below(upper_bound)
		self._map_numbers = map_numbers

	@property
	def keys(self):
		return self._keys_to_values.keys()

	def anonymize_value(self, value, prefix='r'):
		if isinstance(value, str):
			if value in self._values_to_keys:
				return self._values_to_keys[value]
			else:
				key = f'{prefix}{self._count}'
				self._values_to_keys[value] = key
				self._keys_to_values[key] = value
				self._count += 1
				return key
		elif self._map_numbers and isinstance(value, (int, float)):
			return value * self._coefficient + self._constant
		else:
			return value

	def identify_key(self, key):
		if isinstance(key, str):
			return self._keys_to_values[key]
		elif self._map_numbers and isinstance(key, int):
			return int(round((key - self._constant) / self._coefficient))
		elif self._map_numbers and isinstance(key, float):
			return (key - self._constant) / self._coefficient
		else:
			return key

	def anonymize(self, values, shuffle=True, text=None, echo=0):
		"""
		:type values: list or Series
		:type shuffle: bool
		:type echo: int or bool
		:rtype: list
		"""
		echo = max(0, echo)
		values = list(values)
		if shuffle:
			shuffled = values.copy()
			random.shuffle(shuffled)
			[self.anonymize_value(value=x) for x in iterate(shuffled, echo=echo, text='shuffling')]

		return list(ProgressBar.map(
			function=lambda x: self.anonymize_value(value=x),
			iterable=values, echo=echo, text=text or 'anonymizing'
		))

	def identify(self, keys, text=None, echo=0):
		"""
		:type keys: list or Series
		:type echo: bool or int
		:rtype: list
		"""
		echo = max(0, echo)
		keys = list(keys)
		return list(ProgressBar.map(
			function=lambda x: self.identify_key(key=x),
			iterable=keys, echo=echo, text=text or 'identifying'
		))

class SecretKeeper:
	"""
	SecretKeeper is a data anonymizer
	"""
	def __init__(self, map_numbers=False):
		"""
		:type data: DataFrame
		"""
		self._value_anonymizers = {}
		self._name_anonymizer = Anonymizer(map_numbers=False)
		self._map_numbers = map_numbers

	def anonymize_series(self, series, name, shuffle=True, text=None, echo=1):
		"""
		:type series: list or Series
		:type name: str
		:type echo: bool or int
		:rtype: list
		"""
		echo = max(0, echo)
		name = str(name)
		if name not in self._value_anonymizers:
			self._value_anonymizers[name] = Anonymizer(map_numbers=self._map_numbers)

		anonymizer = self._value_anonymizers[name]
		return anonymizer.anonymize(values=series, shuffle=shuffle, text=text, echo=echo)

	def identify_series(self, series, key, text=None, echo=1):
		"""
		:type series: list or Series
		:type key: str
		:type echo: bool or int
		:rtype:
		"""
		echo = max(0, echo)
		key = str(key)
		anonymizer = self._value_anonymizers[key]
		return anonymizer.identify(keys=series, text=text, echo=echo)

	def anonymize(self, data, shuffle=True, echo=1):
		"""
		:type data: DataFrame
		:rtype:  data
		"""
		echo = max(0, echo)
		result = DataFrame()
		for name in data.columns:
			anonymized_name = self._name_anonymizer.anonymize_value(value=str(name), prefix='column_')
			result[anonymized_name] = self.anonymize_series(
				series=data[name], name=anonymized_name, shuffle=shuffle, text=f'anonymizing column {name}', echo=echo
			)

		return result

	def identify(self, data, echo=1):
		"""
		:type data: DataFrame
		:type echo: bool or int
		:rtype:
		"""
		echo = max(0, echo)
		result = DataFrame()
		for anonymized_name in data.columns:
			name = self._name_anonymizer.identify_key(key=anonymized_name)
			result[name] = self.identify_series(
				series=data[anonymized_name], key=anonymized_name, text=f'identifying column {name}', echo=echo
			)
		return result


from setuptools import setup, find_packages


def readme():
	with open('./README.md') as f:
		return f.read()


setup(
	name='riddle',
	version='2020.4.28.1',
	license='MIT',

	author='Idin',
	author_email='py@idin.ca',
	url='https://github.com/idin/riddle',

	keywords='encryption decryption',
	description='Python library for encryption',
	long_description=readme(),
	long_description_content_type='text/markdown',
	classifiers=[
		'Development Status :: 3 - Alpha',
		'Intended Audience :: Developers',
		'Programming Language :: Python :: 3 :: Only',
		'Programming Language :: Python :: 3.6',
		'Programming Language :: Python :: 3.7',
		'Topic :: Software Development :: Libraries :: Python Modules'
	],

	packages=find_packages(exclude=("jupyter_tests", ".idea", ".git")),
	install_requires=['dill', 'disk', 'pandas', 'chronometry', 'slytherin'],
	python_requires='~=3.6',
	zip_safe=True
)



from .encrypt import encrypt as _encrypt
from .encrypt import decrypt as _decrypt
from slytherin.hash import hash_object

import random
import dill


# Riddle encrypts and decrypts objects
class Riddle:
	_ENCRYPTION_CHECK = 'Marvolo'

	# a class that can encrypt or decrypt an object
	def __init__(self, key=None, hash=True, hash_base=64):
		self._hash = hash
		self._hash_base = hash_base
		self._dict = {}
		if key is not None:
			key = str(key)
		self._key = None
		self.unlock(key=key)

	def __getstate__(self):
		return {
			'hash': self._hash,
			'hash_base': self._hash_base,
			'dict': self._dict,
			'key': self._key
		}

	def __setstate__(self, state):
		self._hash = state['hash']
		self._hash_base = state['hash_base']
		self._dict = state['dict']
		self._key = state['key']

	def lock(self):
		self._key = None

	def unlock(self, key):
		if key is None:
			raise ValueError('key cannot be None!')
		if self._hash:
			self._key = self.hash(obj=key, base=self._hash_base)
		else:
			self._key = str(key)

	@property
	def key(self):
		if self._key is None:
			raise KeyError('Riddle is locked!')
		else:
			return self._key

	def get(self, item):
		encrypted_item = self._dict[item]
		return self.decrypt(encrypted_item)

	def set(self, item, obj):
		encrypted_item = self.encrypt(obj=obj)
		self._dict[item] = encrypted_item

	def __getitem__(self, item):
		return self.get(item=item)

	def __setitem__(self, key, value):
		self.set(item=key, obj=value)

	def encrypt(self, obj):
		# obj_plus: first element is for making sure the encryption was successful
		# second element is the actual object
		# third element is to make cracking the encryption harder
		key = self.key
		obj_plus = (
			str(round(random.uniform(0, 1), 3)),
			self._ENCRYPTION_CHECK,
			obj,
			str(round(random.uniform(0, 1), 3))
		)
		clear = dill.dumps(obj_plus, protocol=0)
		decoded = clear.decode(encoding='utf-8')
		return _encrypt(key=key, clear=decoded)

	def decrypt(self, obj):
		key = self.key
		clear_string = _decrypt(key=key, encrypted=obj)
		decrypted_list = dill.loads(clear_string.encode(encoding='utf-8'))
		if decrypted_list[1] != self._ENCRYPTION_CHECK:
			raise KeyError('Riddle: decryption failed!')
		return decrypted_list[2]

	def save(self, x, path):
		encrypted = self.encrypt(x)
		dill.dump(obj=encrypted, file=open(file=path, mode='wb'))

	def load(self, path):
		encrypted = dill.load(file=open(file=path, mode='rb'))
		return self.decrypt(encrypted)

	@staticmethod
	def hash(obj, base=64):
		return hash_object(obj=obj, base=base)


from .Riddle import Riddle

from disk import Box


# Locket is a Box that uses a Riddle to encrypt objects before storing and decrypt them after accessing
class Locket(Box):
	def __init__(self, path, key=None, save_interval_seconds=60):
		super().__init__(path=path, save_interval_seconds=save_interval_seconds)
		self._riddle = Riddle(key=key)

	def __getstate__(self):
		state = super().__getstate__()
		state['riddle'] = self._riddle
		return state

	def __setstate__(self, state):
		super().__setstate__(state=state)
		self._riddle = state['riddle']

	def get(self, name):
		encrypted_item = super().get(name=name)
		return self._riddle.decrypt(obj=encrypted_item)

	def set(self, name, obj):
		encrypted_item = self._riddle.encrypt(obj=obj)
		super().set(name=name, obj=encrypted_item)

	@property
	def items(self):
		self.check()
		encrypted_items = self._dict.items()
		return {key: self._riddle.decrypt(obj=encrypted_value) for key, encrypted_value in encrypted_items}

	def unlock(self, key):
		self._riddle.unlock(key=key)

	def lock(self):
		self._riddle.lock()

